%% Main optimization script (compatible with FEDEM R7.1)
% Written by Johan H. Martens, spring 2014

clc                         % clear command window
clear all                   % clear variables
close all                   % close figures    
rng('shuffle')              % reseed MATLAB random number generator

%% User input
masterfile ='OC4-4lvls-2n.fmm';% Master model file
lvls =      4;              % Height/width ratio of jacket
n =         2;              % Nodes along width of jacket
baseID =    10000;          % Beam ID offset in model
jpa =       7;              % Jacket parent assembly number in master modelfile
pop =       16;             % Population size
conc =      4;              % Number of concurrent processes during fedem analysis
endgen =    100;            % Total number of generations (termination criteria)    
yr =        20;             % Design life [years]
ts =        0.05;           % Timestep for data output
eff_t =     30;             % Effective analysis time [s]
fy =        355;            % Yield limit of steel [MPa]
fatlim =    20;             % Omit fatigue check for stress ranges below fatlim
rho =       7850;           % Steel density [kg/m^3]
ps =        15;             % NOK/kg steel in structure
pins =      15000;          % Price of installing one beam [NOK]
Pb =        0.50;           % Probability of activating initial beam
Pm =        Pb*0.02;        % Initial mutation probability
minPm =     Pb*0.02;        % Minimum mutation probability
maxPm =     Pb*0.20;        % Maximum mutation probability
tresPm =    Pb*0.20;        % Diversity treshold for adapting Pm
maxFit =    100;            % Maximum fitness (optimization goal)
Lc =        11;             % Length of chromosomes [bit]
maxDo =     2.0;            % Maximum allowed diameter in model [m] 
minDo =     0.5;            % Minimum allowed diameter in model [m]
minDratio = 0.8;            % Lower inner/outer diameter ratio boundary
maxDratio = 0.99;           % Upper inner/outer diameter ratio boundary
SPD =       0.01;           % Stiffness proportional damping
MPD =       0;              % Mass proportional damping
cuts =      1;              % Number of cuts in chromosome during crossover
sym =       true;           % Toggle symmetry of all four sides of jacket

%% Defining and preallocating variables
% Model info: [beams (#,node1,node2),Nodes (#, position), Symmetry data]
[B,N,S] = jacket_ftl_creator(n,lvls);   % from jacket funtion
T=N(1:end-8,:);             % Nodes (- 8 nodes in transition piece)
t=length(T);                % Number of nodes
b=length(B);                % Number of beams
totmass=zeros(pop,endgen);  % for total mass of structure
obj=zeros(pop,endgen);      % for value of objective function for all generations
fit=zeros(pop,endgen);      % for value of fitness for all generations
poppos=zeros(pop,endgen);   % for position relative to fitnesses
yield=zeros(1,5,endgen);    % for yield failures
crash=zeros(pop,1,endgen);  % for storing crashed individuals
P=zeros(endgen,3);          % for plot data
fatigue=zeros(pop,endgen);  % for storing individuals failed by fatigue 
gentime=zeros(1,endgen);    % for timing computation time of generations
stresstime=zeros(1,endgen); % computation time of stress analysis
fedemtime=zeros(1,endgen);  % computation time of fedem analysis
writetime=zeros(1,endgen);  % time for writing model files
masterB=zeros(1,size(S,2)); % for master symmetry beams
relfit=zeros(pop,endgen);   % for relative fitness
csp=zeros(b,8,pop,endgen);  % for cross sectional parameters of all beams
nts = eff_t/ts+1;           % number of timesteps from analysis
% cross sectional parameter ID in .fmm model file
sectioninfo = {'AREA = '; 'AREA_MOMENT_IP = '; 'AREA_MOMENT_IY = '; 'AREA_MOMENT_IZ = '; 'HYDRO_DB = '; 'HYDRO_DD = '; 'TUBE_DI = '; 'TUBE_DO = '};

%% Initial cross sectional parametres (creation of first generation)
for j=1:pop                 % population
    nr=1;
    for i = 1:b-8           % beams
        % beam created by probability Pb || cornerbeams can not be removed
        if rand(1) < Pb || any(S{end,1}==i)     
            Do=maxDo*rand(1);               % random value between 0 and maxDo [m]
            while Do < minDo                % make sure Do is valid
                Do=maxDo*rand(1);
            end
            Di=Do*rand(1);                  % random inner diameter
            while Di < minDratio*Do || Di > maxDratio*Do % make sure Di is valid
                Di=Do*rand(1);
            end
        else
            Di=0;
            Do=0;
            rb(nr,j)=i;     % matrix containing id of removed beams
            nr=nr+1;
        end    
        I=pi/64*(Do.^4-Di.^4);
        % parameters [Area Ip Iy Iz HydroDB HydroDD Di Do]
        csp(i,:,j)=[pi/4*(Do.^2-Di.^2) 2*I I I Do Do Di Do];  
    end
end
% if symmetry==true, copy masterbeam parameters to slavebeams
if sym==1
    for i=1:size(S,1)
        masterB(i)=S{i,1}(1,1); % creating list of masterbeams
    end
    clear rb
    for ind=1:pop
        nr=1;
        for i=1:size(S,1)  
            csp(S{i,1}(1,:),:,ind)=repmat(csp(masterB(i),:,ind),[length(S{i,1}(1,:)) 1]);
            if csp(masterB(i),8,ind)==0
                % matrix containing id of removed beams
                rb(nr:(nr+length(S{i,1})-1),ind)=S{i,1};    
                nr=nr+length(S{i,1});
            end
        end
    end
end
            
% Writing parameters of beams inside transition piece for all generations:
% values from OC4 documentation
Do=1.2;
Di=1.160;
I=pi/64*(Do.^4-Di.^4);
csp(end-7:end,:,:,:)=repmat([pi/4*(Do.^2-Di.^2) 2*I I I Do Do Di Do],[8 1 pop endgen]); 

% Finding all beams connected to each node. Will later be used to
% determine if a node has no connecting beams
Att=cell(3,t);              % Preallocating cell array
for tri=1:t                 % For all triads/nodes
    Att{1,tri}=T(tri,1);    % store node number in row 1
    for beam=1:b            % For all beams
        % If current triad has a connected beam
        if B(beam,2)==T(tri,1) || B(beam,3)==T(tri,1)
            % Store beamnumber in cell in row 2
            Att{2,tri}=[Att{2,tri} B(beam,1)];        
        end
    end
end

%% Reading master modelfile into cell array
clear fmmcell
if (exist(masterfile,'file') == 2)      % checking for masterfile
    Fin = fopen(masterfile, 'r');       % Open masterfile for reading
    % Saving file as a cell array, one cell per line
    fmmcell=textscan(Fin,'%s','Delimiter','\n');
else
    fprintf('No master model file!')
    return
end

%% Optimization loop through all generations
for gen=1:endgen
fprintf('\nCommencing generation %d \n\n',gen)
iterationtime=tic;  	% timer of each generation
clear currentmodel
remT=cell(pop,2);    	% Vector will contain triads to be removed
h=1;

%% Creating new model files with updated parameters
fprintf('Writing .fmm''s for generation %d, Individuals: ',gen)
writet=tic;
for p=1:pop
    fprintf('%d ',p)
    readbeams = 0;   	% logicals and counters for reading of .fmm masterfile
    readtriads = 0;
    readsections = 0;
    readjoints = 0;
    readcurves = 0;
    beam_ID = 0;
    % lines to be ignored from masterfile when generating .fmm files  
    del_lns=[];
    % model filename
    currentmodel{p} = sprintf('%s_%03.0f_%03.0f.fmm',masterfile(1:end-4),gen,p);               
    
    %% Identifying nodes without connecting beams
    for tri=1:t                           	% For all nodes
        % Row 3 of Att will contain a counter of matched beams
        Att{3,tri}(1,p)=0;                 
        for remB = 1:size(rb(:,p,gen),1)   	% For all removed beams
            if rb(remB,p,gen)~=0
                % For all beams connected to node
                for attB = 1:length(Att{2,tri})
                    % If match (connected==removed)
                    if Att{2,tri}(1,attB)==rb(remB,p,gen)+baseID
                        % Count + 1 match
                        Att{3,tri}(1,p)=Att{3,tri}(1,p)+1;                                    
                    end
                end
            end
        end
        % Number of connected beams that are removed
        match=Att{3,tri}(1,p);
        % Number of connected beams
        connected=length(Att{2,tri});
        % If all beams connected == number of matches (removed beams)
        if connected==match
            % List triad/node in removal vector remT{p,1}
            remT{p,1}=[remT{p,1} Att{1,tri}(1,1)];         
        end
    end
 
    %% Reading through entire masterfile line by line and customizing
    for ln=1:length(fmmcell{1,1})       
        Lin=fmmcell{1,1}{ln,:}; 	% Get next line of masterfile
        
        %% Beam definitions
        if readbeams == 0                           
            if (strfind(Lin, '!*** Beams ***') > 0) 
               readsections = 0;
               readtriads = 0;
               readbeams = 1;	% Activate beamreading
               readjoints = 0;
               readcurves = 0;
            end
        else
            if (strfind(Lin,sprintf('ID ='))>0)
                for i=1:size(rb,1)
                    % If beam is removed from model
                    if (strfind(Lin,sprintf('ID = %.0f;',rb(i,p,gen)+baseID)) > 0)
                        % checking for correct assembly number
                        if (strfind(fmmcell{1,1}{ln+7,:},sprintf('PARENT_ASSEMBLY = %d',jpa)) > 0)
                            % Delete entry from model file
                            del_lns=[del_lns ln-4:ln+16];                            
                        end
                    end
                end
            % If beam is active -> add correct damping factor
            elseif (strfind(Lin,sprintf('STIF_PROP_DAMP'))>0)
                if (strfind(fmmcell{1,1}{ln+2,:},sprintf('aID: %d',jpa))>0)
                    fmmcell{1,1}{ln,:}=sprintf('STIF_PROP_DAMP = %f;',SPD);
                end
            elseif (strfind(Lin,sprintf('MASS_PROP_DAMP'))>0)
                if (strfind(fmmcell{1,1}{ln+10,:},sprintf('aID: %d',jpa))>0)
                    fmmcell{1,1}{ln,:}=sprintf('MASS_PROP_DAMP = %f;',MPD);
                end
            end
        end
        
        %% Triad (node) definitions
        % only enter triadreading if any triad must be removed
        if ~isempty(remT{p,1})
            % If any triads are to be removed
            if readtriads == 0
                if (strfind(Lin, '!*** Triads ***') > 0)
                   readsections = 0;
                   readtriads = 1;	% Activate triadreading
                   readbeams = 0;
                   readjoints = 0;
                   readcurves = 0;
                end
            else
                for i=1:length(remT{p,1})
                    % If triad is removed from model
                    if (strfind(Lin,sprintf('ID = %.0f;',remT{p,1}(1,i))) > 0)
                        % checking for correct assembly number
                        if (strfind(fmmcell{1,1}{ln+4,:},sprintf('PARENT_ASSEMBLY = %d',jpa)) > 0)
                             % Delete entry from model file
                            del_lns=[del_lns ln-8:ln+5];                           
                        end
                    end
                end
            end
            % If a removed triad is connected to a joint, the joint also
            % has to be removed:
            if readjoints == 0                           
                if (strfind(Lin, '!*** Joints ***') > 0) 
                   readsections = 0;
                   readtriads = 0;
                   readbeams = 0;
                   readjoints = 1; % Activate jointreading
                   readcurves = 0;
                end
            else
                for i=1:length(remT{p,1})
                    % If triad is removed from model
                    if (strfind(Lin,sprintf('SLAVE_TRIAD = aID: %.0f uID: %.0f;',jpa,remT{p,1}(1,i))) > 0)
                        % Delete corresponding joint
                        del_lns=[del_lns ln-16:ln+3];                           
                    end
                end
            end
        end
        
        %% Beam cross section definitions
        if readsections == 0
           if (strfind(Lin, '!*** Beam cross sections ***') > 0)
               readsections = 1; 	% Activate reading of beam cross sections
               readtriads = 0;
               readbeams = 0;
               readjoints = 0;
               readcurves = 0;
           end
        else           
            if (strfind(Lin,'ID')>0)% if keyword ID is found
                for i=1:b           % cycling through all beams in model
                    if (strfind(Lin,sprintf('ID = %.0f;',i+baseID))>0)
                        % checking for correct material and parent assembly
                        if (strfind(fmmcell{1,1}{ln+1,:},sprintf('MATERIAL = aID: %.0f uID: 1;',jpa)) > 0) 
                            if beam_ID==0
                                % If beam is removed from model
                                if any(rb(:,p,gen)==i)
                                     % Delete entry from model file
                                    del_lns=[del_lns ln-23:ln+8];      
                                    break
                                else
                                    % if beam is active in model
                                    beam_ID=i; % store ID                          
                                    break
                                end
                            end
                        end
                    end
                end
            end
            if beam_ID > 0                      % If beam is active
                for var_ln=ln-24:ln+7           % For relevant lines
                    Lin=fmmcell{1,1}{var_ln,:};	% Extract line
                    for k=1:8                 	% Cycle through parameters
                        % Find parameter in current line
                        if (strfind(Lin,sectioninfo{k}) > 0)
                            % Replacing parameters with updated ones
                            fmmcell{1,1}{var_ln,:}=sprintf('%s%f;',sectioninfo{k},csp(beam_ID,k,p,gen));
                            break
                        end
                    end
                end
            end
            beam_ID=0;
        end
        
        %% Curve export definitions
        if readcurves == 0                           
            if (strfind(Lin, '!*** Curves ***') > 0) 
               readsections = 0;
               readtriads = 0;
               readbeams = 0;           
               readjoints = 0;
               readcurves = 1;  % Activate curvereading
            end
        else  % finding result object
            if (strfind(Lin,sprintf('Y_AXIS_RESULT_OBJECT = aID: %d',jpa)) > 0)
                for i=1:b % finding corresponding beam
                    if (strfind(Lin,sprintf('uID: %d',i+baseID)) > 0)
                        % if current curve corresponds to a removed beam
                        if any(rb(:,p,gen)==i)
                            % Delete entry from model file
                            del_lns=[del_lns ln-42:ln+5];       
                            break
                        end
                    end
                end
            end
        end
    end
    
    %% Making folders and writing model file of current individual
    inddir=sprintf('Ind_%03.0f_%03.0f',gen,p);
    mkdir(inddir)
    modelpath=sprintf('%s\\%s',inddir,currentmodel{p});
    % Create model file for current iteration
    Fout = fopen(modelpath,'w');                        
    for ln=1:length(fmmcell{1,1})   % For all lines in masterfile   
        if ~any(del_lns==ln)        % As long as line is not marked to be deleted
            % Copy line to new modelfile
            fprintf(Fout,'%s\n',fmmcell{1,1}{ln,:});    
        end
    end
    % Copy result directory and transition piece defintion to new individual
    cf_source      = [masterfile(1:end-4) '_RDB'];
    cf_destination = [inddir '\' currentmodel{p}(1:end-4) '_RDB'];
    copyfile(cf_source,cf_destination);
end
writetime(gen)=toc(writet); % model file writing timer
fprintf('\nDone!\n\n')
fclose('all');
 
%% Run model files of current generation in FEDEM
fedemt=tic;
for runs=1:conc:pop
    indstr=sprintf('%d ',[runs:runs+conc-1]);
    fprintf('Running FEDEM. Generation: %d, Individuals: %s\n',gen,indstr);
    % Parallel for loop for of "conc" models for faster computation
    parfor p=runs:runs+conc-1 
        inddir=sprintf('Ind_%03.0f_%03.0f',gen,p);
        modelpath=sprintf('%s\\%s',inddir,currentmodel{p})
        PSrun = sprintf('powershell -inputformat none fedem -f %s -solve dynamics',modelpath);
        system(PSrun);
    end
    fprintf('Done!\n\n')
end
fedemtime(gen)=toc(fedemt); % timer of fedem analysis

%% Extracting timeseries results from analysis and evaluating damage
stresst=tic;
cfail=1;
yfail=1;
fprintf('Importing forcedata from analysis, converting to MPa\n and checking for yielding and fatigue damage\n\n')
for ind=1:pop  	% for all individuals 
    file = sprintf('Ind_%03.0f_%03.0f\\timeseries.asc',gen,ind);
    if (exist(file,'file') == 2)
        % loading forcedata from file into the struct timeforces
        timeforces=importdata(file,'\t',7); 
        timeforces.data(:,1)=[];      	% deleting timestep column
        % if correct number of datapoints
        if size(timeforces.data,1)==nts	
            for beam=1:b-8	% for all beams (except 8 in transition piece)
                % if beam is not removed and individual has not failed
                if ~any(rb(:,ind,gen)==beam) && fatigue(ind,gen)==0 && ~any(yield(:,1,gen)==ind) && ~any(crash(:,1,gen)==ind)             
                    Do= csp(beam,8,ind,gen);    % Outer diameter
                    A = csp(beam,1,ind,gen);    % Area
                    I = csp(beam,3,ind,gen);    % Moment of inertia
                    % number of skipped beams so far
                    ns = sum(rb(:,ind,gen)<beam) - sum(rb(:,ind,gen)==0);
                    % number of active beams
                    na = b - sum(rb(:,ind,gen)>0);                        
                    for End=1:2 	% for both ends of beam
                        % COLLAPSING CODE FOR READABILITY
% Extracting forces of current beam end, calculating stresses and converting to MPa
if End==1    
    timestresses=timeforces.data(:,[beam beam+na beam+2*na]-ns).*repmat([1/A (Do/2)/I (Do/2)/I],nts,1)./(1000^2);
elseif End==2
    timestresses=timeforces.data(:,3*na+[beam beam+na beam+2*na]-ns).*repmat([1/A (Do/2)/I (Do/2)/I],nts,1)./(1000^2);
end
% Checking for empty datasets (ISSUE WITH EMPTY CURVE EXPORTS FROM FEDEM!)
if all(timestresses==0)
    % Delete individuals with totally unloaded members
    fprintf('Warning: empty matrix of stresses extracted from individual %d, beam %d\n',ind,beam) 
    fprintf('Individual discarded from optimization\n')
    crash(cfail,1,gen)=ind;	% storing crashed individuals
    cfail=cfail+1;
    break
end
sigx=timestresses(:,1);     % Stress from axial force
sigy=timestresses(:,2);     % Stress from moment about Y-axis
sigz=timestresses(:,3);     % Stress from moment about Z-axis
% Calculating stresses in 8 spots around every beam end
% according to DNV-RP-C203 s.32 (SCF=1)
clear sig
sig(:,1)=sigx+sigy;
sig(:,2)=0.5*sigx+sqrt(2)/2*sigy-sqrt(2)/2*sigz;
sig(:,3)=sigx-sigz;
sig(:,4)=0.5*sigx-sqrt(2)/2*sigy-sqrt(2)/2*sigz;
sig(:,5)=sigx-sigy;
sig(:,6)=0.5*sigx-sqrt(2)/2*sigy+sqrt(2)/2*sigz;
sig(:,7)=sigx+sigz;
sig(:,8)=0.5*sigx+sqrt(2)/2*sigy+sqrt(2)/2*sigz;

%% Checking for yielding and fatigue damage
maxstress=max(max(abs(sig)));   % maximum stress in current section
if maxstress > fy
    % for first yield current in generation
    if yfail == 1
        yield(yfail,1:3,gen)=[ind beam+baseID maxstress];
        yfail=yfail+1;
        fprintf('Individual %d failed by yielding at %0.f MPa (beam: %d)\n',ind,maxstress,beam+baseID)
        break
    % if a new individual yields
    elseif ind ~= yield(yfail-1,1,gen)
        yield(yfail,1:3,gen)=[ind beam+baseID maxstress];
        yfail=yfail+1;
        fprintf('Individual %d failed by yielding at %0.f MPa (beam: %d)\n',ind,maxstress,beam+baseID)
        break
    end
end
% only do fatigue check if highest stress cycle is above fatlim
if max(range(sig,1)) > fatlim
     % run fatigue analysis in external function, 
     % fatiguefail returns fatigue(ind,gen)=1 for failure
    fatigue(ind,gen)=fatiguefail(sig,yr,eff_t);    
    if fatigue(ind,gen)==1 
        fprintf('Individual %d failed by fatigue (beam ID: %d)\n',ind,beam+baseID)
        break
    end
end
                        % BRINGING CODE UP TO CORRECT LEVEL AGAIN
                    end
                end
            end
        else % not enough datapoints (premature termination of analysis)
            fprintf('Corrupt timeseries data for individual %d\n',ind)
            crash(cfail,1,gen)=ind;	% storing crashed individual
            cfail=cfail+1;
        end
    else % no data (totally failed analysis)
        fprintf('No timeseries data for individual %d\n',ind)
        crash(cfail,1,gen)=ind;     % storing crashed individual
        cfail=cfail+1;                      
    end
end
stresstime(gen)=toc(stresst);
fprintf('Done!\n\n')

%% Calculating weight of structures
totmass(:,gen)=zeros(pop,1);
% looping through population of current generation
for ind=1:pop                                   
    for j=1:b-8                                 % cycling through beams
        if ~any(rb(:,ind,gen)==j)               % exclude removed beams
            % extracting triad coordinates [x1 y1 z1; x2 y2 z2]
            pos=[N(N==B(j,2),2:4); N(N==B(j,3),2:4)]; 
            dx=pos(2,1)-pos(1,1);               % length along axes
            dy=pos(2,2)-pos(1,2);
            dz=pos(2,3)-pos(1,3);
            len=sqrt(dx^2+dy^2+dz^2);           % length of beam j
            vol=len*csp(j,1,ind,gen);           % volume of beam j [m^3]
            % mass of beam j added to totmass(pop x gen) [kg]
            totmass(ind,gen)=totmass(ind,gen)+vol*rho;              
        end
    end
end

%% Calculating objective function/fitness and updating leadertable
for i=1:pop
    % price of structure (estimate in NOK), the more removed beams the cheaper
    obj(i,gen)=totmass(i,gen)*ps+(b-sum(rb(:,i,gen)>0))*pins;
    % fitness = constant - objective value, (higher = better)
    fit(i,gen)=maxFit-obj(i,gen)/1E6;                                    
end
% sorting by fitness and pairing with correct population numnber
[fit(:,gen), poppos(:,gen)]=sort(fit(:,gen),'descend');
% result of current generation before yield and fatigue
genres=[fit(:,gen), poppos(:,gen)];                         
nr=0;           % counter
clear surv      % preparing variable of survivors
for i=1:pop
    if ~any(genres(i,2)==crash(:,1,gen))      	% if individual did not crash
        if ~any(genres(i,2)==yield(:,1,gen))  	% if individual did not yield 
            if fatigue(genres(i,2),gen)==0     	% and not fail by fatigue
                nr=nr+1;                       	% add to number of survivors
                surv(nr,:)=genres(i,:);        	% store survivors
            end
        end
    end
end

ncrash=sum(crash(:,1,gen)>0);   % number of crashed individuals
nfat=sum(fatigue(:,gen)>0);     % number of failures by fatigue  
nyield=sum(yield(:,1,gen)>0);   % number of failures by yeilding
fprintf('%d of %d individuals survived generation %d!\n',nr,pop,gen)
if nr~=pop
    fprintf('%d failed by fatigue, %d crashed and %d yielded\n',nfat,ncrash,nyield)
end
nosurv=0;
if ~exist('surv','var')
    fprintf('\nNo individuals from generation %d survived\n',gen)
    nosurv=1;
    if ~exist('leaders','var')
        fprintf('\nAll initial designs failed! Optimization terminated. \n')
        return
    end
end
% if there are survivors, update leadertable
if nosurv==0                                               
    if gen==1
        % for first generation all survivors are leaders
        templead=[surv ones(size(surv,1),1)*gen];           
    else
        % adding result of current generation (>1) to temporary leader table
        templead=[leaders(:,:,gen-1);surv ones(size(surv,1),1)*gen]; 
    end
    % sorting list containing old leaders and new candidates
    templead=sortrows(templead,1);                          
    % inverting list (best fitness on top)
    temp=templead;
    for i=1:size(templead,1)                             
        templead(i,:)=temp(end+1-i,:);
    end
    % creating new leadertable
    if length(templead) > pop
        % saving sorted top designs as new leaders
        leaders(1:pop,:,gen)=templead(1:pop,:);             
    else
        leaders(1:size(templead,1),1:3,gen)=templead;
    end
    P(gen,2)=surv(1,1);                	% best design within current generation
else                                  	% if there are no survivors
    leaders(:,:,gen)=leaders(:,:,gen-1);% copy leadertable from previous generation
    P(gen,2)=NaN;                      	% no best design in current generation
end
% If only one survivor of generation 1, then terminate
if size(leaders,1)==1 && gen==1
    fprintf('\nOnly one surviving design of first generation! Optimization terminated. \n')
    return
end

%% Plotting optimization evolution graph
P(gen,1)=leaders(1,1,gen);          % overall leader in generation i
P(gen,3)=mean(leaders(:,1,gen));    % mean of leadertable in generation i
P(gen,4)=pop-nr;                    % number of casualties in current generation
close all
figure
[ax, h1, h2] = plotyy(1:gen,P(1:gen,1:3),1:gen,P(1:gen,4),'plot','bar');
set(h1(1),'LineWidth',3)
set(h1(2),'LineWidth',1.5)
set(h1(3),'linestyle','--','color',[.3 .3 .3],'LineWidth',2.5)
set(h2, 'FaceColor', [.6 0 0]);
hleg=legend('Leading Design','Generation Winner','Mating Pool Mean');
set(hleg,'FontSize',24,'Interpreter','latex','position',[-0.04 0.75 0.4 0.3]);
xlabel('\textbf{Generation}','FontSize',26,'Interpreter','latex')
set(get(ax(1), 'Ylabel'), 'String','\textbf{Fitness}','FontSize',26,'Interpreter','latex');
set(ax(1),'Xlim',[1 gen+2])
set(ax(1),'YLim',[80 100],'YTick',[80:5:100],'fontsize',20);
set(ax(2),'Xlim',[1 gen+2],'XTick',[0:10:gen])
y2lab=get(ax(2), 'Ylabel');
set(y2lab,'String','\textbf{Casualties}','color',[.6 0 0],'FontSize',26,'Interpreter','latex');
set(y2lab,'Units','Normalized','Position',[1.03 .2 0])
set(ax(2),'YLim',[0 pop*3],'YTick',[0:2:pop],'YColor',[.6 0 0],'fontsize',20);
grid on
grid minor
set(gcf,'Units','normal')
set(gca,'Position',[.05 .07 .9 .91])
drawnow;
save('plotdata.txt','P','-ascii')

%% printing leadertable to command window
fprintf('\nLeading designs at end of generation %d \n [Fitness Individual Generation]\n',gen)
disp(leaders(:,:,gen))

%% Calculating relative fitnesses and creating mating pool
fprintf('\nPreforming crossover and mutation\n')
% Calculating relative fitness for use in weighted roulette wheel
% scaling fitness such that worst leader has zero fitness (will not pass on genome)
relfit(1:size(leaders,1),gen)=leaders(:,1,gen)-min(leaders(:,1,gen));
% making sum(relfit) = 1 -> relfit conatins probabilities of becoming parents
relfit(:,gen)=relfit(:,gen)/sum(relfit(:,gen));                         
% If sum(relfit(:,gen))=0 -> NaN will be generated, if so, use relfit and
% leadertable of previous generation
if any(isnan(relfit(:,gen)))
    relfit(:,gen)=relfit(:,gen-1);
    leaders(:,:,gen)=leaders(:,:,gen-1);
    fprintf('\n Too low diversity in leadertable, reusing leadertable of generation %d \n',gen-1)
end
n_pool=size(leaders(:,:,gen),1);    % number of parents
parents=cell(2,3);                  % clearing parent matrix
matingpool=cell(n_pool,2);          % clearing mating pool
% creating mating pool (binary chromosomes)
if sym==0
    for i=1:n_pool
        % generation and population number of leader i
        popgen=[leaders(i,2,gen) leaders(i,3,gen)];
        % converting Di (in mm) to binary (Lc bits)
        matingpool{i,1}=dec2bin(1000*csp(1:end-8,7,popgen(1),popgen(2)),Lc);
        % converting Do (in mm) to binary (Lc bits)
        matingpool{i,2}=dec2bin(1000*csp(1:end-8,8,popgen(1),popgen(2)),Lc);    
    end
% if symmetry, then only masterbeams are design variables
elseif sym==1
    for i=1:n_pool                      
        popgen=[leaders(i,2,gen) leaders(i,3,gen)];
        matingpool{i,1}=dec2bin(1000*csp(masterB,7,popgen(1),popgen(2)),Lc);  
        matingpool{i,2}=dec2bin(1000*csp(masterB,8,popgen(1),popgen(2)),Lc);   
    end
end

%% Adaptive mutation probability
% number of equal genes between best and worst individual in mating pool
eqg=sum(sum(matingpool{1,1}==matingpool{end,1}));
% number of genes in total
numg=size(matingpool{1,1},1)*Lc;
% gene diversity
div=(numg-eqg)/numg;                                
% adjust mutation probaility
if div > tresPm
    if Pm > minPm
        Pm = Pm - Pb*0.01;
    end
else
    if Pm < maxPm
        Pm = Pm + Pb*0.01;
    end
end

%% Breeding (crossover and mutation) 
clear children
children=cell(n_pool,2);
c=1;                    % children counter
while c < pop           % making "pop" new children two at a time
    parent=cell(2,3);
    rndm=rand(1);       % selecting first parent by weighted roulette wheel
    for k=1:n_pool
        if sum(relfit(1:k,gen))>=rndm                     
            parent{1,1}=matingpool{k,1}; % Extracting beam Di's of first parent
            parent{1,2}=matingpool{k,2}; % Extracting beam Do's of first parent
            parent{1,3}=k;   	% Storing parent ID to avoid identical parents
            break
        end
    end
    rndm=rand(1);       % selecting second parent by weighted roulette wheel
    for k=1:n_pool
        if sum(relfit(1:k,gen))>=rndm
            % Making sure parents are not identical 
            if k~=parent{1,3}                    
                parent{2,1}=matingpool{k,1};% Extracting beam Di's of second parent
                parent{2,2}=matingpool{k,2};% Extracting beam Do's of second parent
                parent{2,3}=k;
                break
            end
        end
    end
    % Preforming crossover for all design parameters (chromosomes) 
    % Establishing cut positions for current parents
    cpos=zeros(1,cuts);
    for cut=1:cuts
        if cut==1
            % finding first cut position (random)
            cpos(cut)=ceil(rand(1)*(Lc-cuts));          
        else
            % finding consecutive cut positions
            cpos(cut)=ceil(rand(1)*(Lc-cuts+cut-1));
            % checking that position is valid
            while cpos(cut) <= cpos(cut-1)   	
               cpos(cut)=ceil(rand(1)*(Lc-cuts+cut-1));
            end
        end
    end
    % Breeding two children from two parents by crossover:    
    for j=1:size(parent{1,1},1)	% beams
        cross=0;                % 0: parent 1 to child 1, 1: parent 2 to child 1
        for cut=1:cuts+1
            % First cut
            if cut==1
                % Di beam j child c
                children{c  ,1}(j,1:cpos(cut))=parent{1+cross,1}(j,1:cpos(cut));
                % Do beam j child c
                children{c  ,2}(j,1:cpos(cut))=parent{1+cross,2}(j,1:cpos(cut));
                % Di beam j child c+1
                children{c+1,1}(j,1:cpos(cut))=parent{2-cross,1}(j,1:cpos(cut));
                % Do beam j child c+1
                children{c+1,2}(j,1:cpos(cut))=parent{2-cross,2}(j,1:cpos(cut));
            % Intermediate cut(s)
            elseif cut < cuts+1                                                    
                children{c  ,1}(j,cpos(cut-1)+1:cpos(cut))=parent{1+cross,1}(j,cpos(cut-1)+1:cpos(cut));      
                children{c  ,2}(j,cpos(cut-1)+1:cpos(cut))=parent{1+cross,2}(j,cpos(cut-1)+1:cpos(cut));
                children{c+1,1}(j,cpos(cut-1)+1:cpos(cut))=parent{2-cross,1}(j,cpos(cut-1)+1:cpos(cut));   
                children{c+1,2}(j,cpos(cut-1)+1:cpos(cut))=parent{2-cross,2}(j,cpos(cut-1)+1:cpos(cut));
            % Last cut
            elseif cut==cuts+1                                                      
                children{c  ,1}(j,cpos(cut-1)+1:Lc)=parent{1+cross,1}(j,cpos(cut-1)+1:Lc);      
                children{c  ,2}(j,cpos(cut-1)+1:Lc)=parent{1+cross,2}(j,cpos(cut-1)+1:Lc);
                children{c+1,1}(j,cpos(cut-1)+1:Lc)=parent{2-cross,1}(j,cpos(cut-1)+1:Lc);   
                children{c+1,2}(j,cpos(cut-1)+1:Lc)=parent{2-cross,2}(j,cpos(cut-1)+1:Lc);
            end
            % Switching crossing factor between cuts
            if cross==0     
                cross=1;
            else
                cross=0;
            end
        end   
        % Mutation of Di and Do for beam j of both children
        for bit=1:Lc    % for all bits in chromosome
            % Mutation criteria
            if rand(1) <= Pm
                % child c, Di mutation
                switch children{c,1}(j,bit)       
                    case '0'
                        children{c,1}(j,bit)='1';
                    case '1'
                        children{c,1}(j,bit)='0';
                end
                % child c, Do mutation
                switch children{c,2}(j,bit)         
                    case '0'
                        children{c,2}(j,bit)='1';
                    case '1'
                        children{c,2}(j,bit)='0';
                end
                % child c+1, Di mutation
                switch children{c+1,1}(j,bit)              
                    case '0'
                        children{c+1,1}(j,bit)='1';
                    case '1'
                        children{c+1,1}(j,bit)='0';
                end
                % child c+1, Do mutation
                switch children{c+1,2}(j,bit)        
                    case '0'
                        children{c+1,2}(j,bit)='1';
                    case '1'
                        children{c+1,2}(j,bit)='0';
                end
            end
        end
    end
    c=c+2;	% Updating children counter, 2 children produced
end
% converting children Di and Do chromosomes back to decimal and meters
if sym==0
    for i=1:pop
        csp(1:end-8,7,i,gen+1)=bin2dec(children{i,1}(:,1:Lc))/1000; 
        csp(1:end-8,8,i,gen+1)=bin2dec(children{i,2}(:,1:Lc))/1000; 
    end
elseif sym==1
    for i=1:pop
        csp(masterB,7,i,gen+1)=bin2dec(children{i,1}(:,1:Lc))/1000; 
        csp(masterB,8,i,gen+1)=bin2dec(children{i,2}(:,1:Lc))/1000; 
    end
end

%% Producing new set of valid design parameters from breeding result
for j=1:pop
    nr=1;       % counter for removed beams
    if sym==0   % for unsymmetric case
        for i=1:b-8
            Do=csp(i,8,j,gen+1);
            if Do < minDo       	% remove beams with Do < minDo
                Do=0;
                rb(nr,j,gen+1)=i; 	% matrix containing id of removed beams
                nr=nr+1;
            elseif Do > maxDo    	% limit diameter to maxDo                 
                Do=maxDo;
            end          
            Di=csp(i,7,j,gen+1);
            if Do~=0                % if beam is not removed
                % make sure Di is valid
                if Di < minDratio*Do || Di > maxDratio*Do
                    while Di < minDratio*Do || Di > maxDratio*Do 
                        Di=Do*rand(1);
                    end
                end
            else % if beam is removed
                Di=0;
            end
            I=pi/64*(Do.^4-Di.^4);
            % adding [Area Ip Iy Iz HydroDB HydroDD Di Do] to csp for next gen
            csp(i,1:8,j,gen+1)=[pi/4*(Do.^2-Di.^2) 2*I I I Do Do Di Do];   
        end
    % if symmerty, then masterbeams decide parameters of slavebeams
    elseif sym==1
        for i=1:length(masterB)
            Do=csp(masterB(i),8,j,gen+1);
            if Do < minDo      	% remove beams with Do < minDo
                Do=0;
                % matrix containing id of removed beams
                rb(nr:(nr+length(S{i,1})-1),j,gen+1)=S{i,1};    
                nr=nr+length(S{i,1});
                % corner beams cannot be removed -> generate random cross section!
                if any(S{end,1}==masterB(i))
                    % make sure Do is valid
                    while Do < minDo            
                        Do=maxDo*rand(1);
                    end
                end
            % limit diameter to maxDo 
            elseif Do > maxDo                               
                Do=maxDo;
            end          
            Di=csp(masterB(i),7,j,gen+1);
            if Do~=0    	% if beam is not removed
                % make sure Di is valid
                if Di < minDratio*Do || Di > maxDratio*Do
                    while Di < minDratio*Do || Di > maxDratio*Do 
                        Di=Do*rand(1);
                    end
                end
            else
                Di=0;
            end
            I=pi/64*(Do.^4-Di.^4);
            % filling csp matrix with symmetrical beam properties
            csp(S{i,1},1:8,j,gen+1)=repmat([pi/4*(Do.^2-Di.^2) 2*I I I Do Do Di Do],[length(S{i,1}(1,:)) 1 1 1]);
        end
    end
end
fprintf('Done!\n')

%% Removing 2 generations old RDB folders to free disc space
% does not remove .fmm (can be rerun to bring back RDB)
if gen > 2
    fprintf('\nDeleting RDB folders for generation %d\n',gen-2)
    for ind=1:pop
        try
            inddir=sprintf('Ind_%03.0f_%03.0f',gen-2,ind);
            model = sprintf('%s_%03.0f_%03.0f.fmm',masterfile(1:end-4),gen-2,ind);
            RDBpath=sprintf('%s\\%s',inddir,model);          
            rmdir(sprintf('%s_RDB',RDBpath(1:end-4)),'s')
        catch
            fprintf('Could not delete RDB folder for individual %d\n',ind)
        end
    end
    fprintf('Done!\n')
end
% storing elapsed generation time
gentime(gen)=toc(iterationtime);


% Generation completed, go to top to commence next generaion
end

%% Optimization successfully completed, present total time used!
fclose('all');
tottime=sum(gentime)/60;
totstress=sum(stresstime)/60;
totfedem=sum(fedemtime)/60;
totwrite=sum(writetime)/60;
fprintf('\nTotal fedem analysis time: %d h %d min!\n',floor(totfedem/60),round(rem(totfedem,60)))
fprintf('\nTotal stress analysis time: %d h %d min!\n',floor(totstress/60),round(rem(totstress,60)))
fprintf('\nTotal model writing time: %d h %d min!\n',floor(totwrite/60),round(rem(totwrite,60)))
fprintf('\nScript successfully completed in: %d h %d min!\n',floor(tottime/60),round(rem(tottime,60)))
